close all
clear all

im = imresize(imread('cat.jpg'), 0.5);      % read img
[ny, nx, nc] = size(im);                    % size of img
f = double(im(:)) / 255.; 
f = f + randn(size(f))*.1;                  % noisy observation
m = ones(size(f));
                          
nabla = spmat_gradient2d(nx,ny,nc);         % discrete gradient

alpha = 0.1;                               % regularization weight

max_iter = 500;                             % max iteration
en_primal = @(u) 0.5 * sum( (u-f).^2 ) + alpha * sum(abs(nabla * u));
                                            % primal energy evaluation
en_dual = @(p) -(0.5 * sum( (nabla' * p).^2 ) - sum(f .* (nabla' * p)));
                                            % dual energy evaluation
                                            
method = 'fbs';                             % choose solver


%% forward-backward splitting
if strcmp(method, 'fbs')
    u = f;
    p = zeros(ny*nx*nc*2, 1);
%     tau = 1/(normest(nabla).^2);
    tau = 1/8;
    
    energy_fbs = zeros(max_iter+1, 1);
    gap_fbs = zeros(max_iter+1, 1);
    energy_fbs(1)= en_primal(u);
    gap_fbs(1) = en_primal(u) - en_dual(p);

    for k=1:max_iter
        if mod(k,10) == 0
            fprintf('iter %d: primal-ener=%.6g, duality-gap=%.3e\n', ...
                k, en_primal(u), en_primal(u) - en_dual(p));
        end
        
        grad = nabla * (nabla' * p - f);
        p = p - tau * grad;
        
        pproj = p; 
        pproj(pproj > alpha) = alpha;
        pproj(pproj < -alpha) = -alpha;
        p = pproj(:);
        
        u = f - (nabla' * p);
        
        energy_fbs(k+1)= en_primal(u);
        gap_fbs(k+1) = en_primal(u) - en_dual(p);
    end
    
    % display result
    figure(100)
    subplot(221)
    imshow(reshape(f, [ny nx nc]));
    title('observation')
    subplot(222)
    imshow(reshape(u, [ny nx nc]));
    title('denoised image')
    subplot(223)
    plot(0:k,energy_fbs(1:k+1))
    axis([0,max_iter,0,max(energy_fbs)*1.5])
    hold on
    title('primal energy')
    subplot(224)
    semilogy(0:k,gap_fbs(1:k+1))
    axis([0,max_iter,0,max(gap_fbs)*2])
    hold on
    title('duality gap')
end


%% ADMM
if strcmp(method, 'admm')

    u = f;
    v = zeros(ny*nx*nc*2, 1);
    p = zeros(ny*nx*nc*2, 1);

    tau = 25;
    
    energy_admm = zeros(max_iter+1, 1);
    gap_admm = zeros(max_iter+1, 1);
    energy_admm(1)= en_primal(u);
    gap_admm(1) = en_primal(u) - en_dual(p);
    
    M = speye(size(nabla,2)) + tau * nabla' * nabla;
    for k=1:max_iter        
        if mod(k,10) == 0
            fprintf('iter %d: primal-ener=%.6g, duality-gap=%.3e\n', ...
                k, en_primal(u), en_primal(u) - en_dual(p));
        end

        % u-update
        u = M \ (f - nabla' * (p - tau * v));
        
        % v-update
        prox_arg = nabla * u + p / tau;
        prox_step = alpha / tau;
        v = max(0, prox_arg - prox_step) - max(0, -prox_arg - prox_step);
        
        % p-update        
        p = p + tau * (nabla * u - v);
        
        energy_admm(k+1)= en_primal(u);
        gap_admm(k+1) = en_primal(u) - en_dual(p);
    end
    
    % display result
    figure(100)
    subplot(221)
    imshow(reshape(f, [ny nx nc]));
    title('observation')
    subplot(222)
    imshow(reshape(u, [ny nx nc]));
    title('denoised image')
    subplot(223)
    plot(0:k,energy_admm(1:k+1))
    axis([0,max_iter,0,max(energy_admm)*1.5])
    hold on
    title('primal energy')
    subplot(224)
    semilogy(0:k,gap_admm(1:k+1))
    axis([0,max_iter,0,max(gap_admm)*2])
    hold on
    title('duality gap')
end


%% PDHG
if strcmp(method, 'pdhg')
    
    u = f;
    p = zeros(ny*nx*nc*2, 1);

    tau = 0.25;
    sigma = 0.5;
    
    energy_pdhg = zeros(max_iter+1, 1);
    gap_pdhg = zeros(max_iter+1, 1);
    energy_pdhg(1)= en_primal(u);
    gap_pdhg(1) = en_primal(u) - en_dual(p);
    
    for k=1:max_iter        
        if mod(k,10) == 0
            fprintf('iter %d: primal-ener=%.6g, duality-gap=%.3e\n', ...
                k, en_primal(u), en_primal(u) - en_dual(p));
        end

        u_tilde = u - tau * nabla' * p;
        
        u_prev = u;
        u = (u_tilde + tau * f) / (1 + tau);
        
        p_tilde = p + sigma * nabla * (2 * u - u_prev);

        pproj = p_tilde; 
        pproj(pproj > alpha) = alpha;
        pproj(pproj < -alpha) = -alpha;
        p = pproj(:);
        
        energy_pdhg(k+1)= en_primal(u);
        gap_pdhg(k+1) = en_primal(u) - en_dual(p);
    end
    
    % display result
    figure(100)
    subplot(221)
    imshow(reshape(f, [ny nx nc]));
    title('observation')
    subplot(222)
    imshow(reshape(u, [ny nx nc]));
    title('denoised image')
    subplot(223)
    plot(0:k,energy_pdhg(1:k+1))
    axis([0,max_iter,0,max(energy_pdhg)*1.5])
    hold on
    title('primal energy')
    subplot(224)
    semilogy(0:k,gap_pdhg(1:k+1))
    axis([0,max_iter,0,max(gap_pdhg)*2])
    hold on
    title('duality gap')
end

